/* audio file ξɽ
 * 10/15 2000
 * shige
 * aufinfo.c
 *
 * $Id: aufinfo.c,v 1.4 2000-10-22 12:02:12+09 shige Exp $
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sndio.h"
#include "wavio.h"
#include "error.h"

typedef struct {
    char id[4]; /* chunk id */
    long int size; /* chunk size */
    long int start; /*  offset */
} Chunk;

#ifdef USEMALLOC
typedef struct chunklist {
    Chunk *p;
    struct chunklist *next;
} Chunklist;
#endif /* USEMALLOC */

#define sto4bytebigi(s) ((s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3])
#define sto4bytelittlei(s) ((s[3]<<24)|(s[2]<<16)|(s[1]<<8)|s[0])
#define sto2bytelittlei(s) ((s[1]<<8)|s[0])
#ifdef USEMALLOC
# define ChnkAlloc() ((Chunk *)malloc(sizeof(Chunk)))
# define ChnkListAlloc() ((Chunklist *)malloc(sizeof(Chunklist)))
#else
# define Maxchunks 20
#endif /* USEMALLOC */

enum { BlockSize=0, DataSize=1, DataOffset=2,
       allparams=3 };

/* ####### prototype declaration ####### */
int main(int argc,char *argv[]);
int getarg(int argc,char *argv[],int *Mode,int param[]);
long int getfilesize(FILE *file);
#ifdef USEMALLOC
int getWAVinfo(FILE *file,WAVdata *wd,Chunklist **Cl);
Chunklist *addchunklist(Chunklist *cl,Chunk *ch);
void Clclear(Chunklist *cl);
void putlongwav(char *fname,long int fsize,WAVdata *wd,Chunklist *cl,
		int param[]);
void putchunklist(Chunklist *cl);
#else
int getWAVinfo(FILE *file,WAVdata *wd,Chunk *Ch);
void putlongwav(char *fname,long int fsize,WAVdata *wd,Chunk *ch,
		int chunks,int param[]);
void putchunklist(Chunk *ch,int chunks);
#endif /* USEMALLOC */
void putlongau(char *fname,long int fsize,Snddata *sd,int param[]);
void putdefaultheader(int param[]);
void putwav(char *fname,long int fsize,WAVdata *wd,int param[]);
void putau(char *fname,long int fsize,Snddata *sd,int param[]);
void putwavinfo0(int fmtchunksize,WAVdata *wd);
void usage(void);

/* ##### main routine ##### */
int main(int argc,char *argv[])
{
    char magic[4];
    int mode=0,param[allparams];
    int j,k=0,ret;
    long int totalfilesize;
    FILE *file;
    WAVdata wd;
    Snddata sd;
#ifdef USEMALLOC
    Chunklist *cl=NULL;
#else
    Chunk ch[Maxchunks];
#endif /* USEMALLOC */
    
    if((j=getarg(argc,argv,&mode,param))<0
       ||(mode!=0 && mode!=1)){
	usage(); return Exit_getarg;
    }
    for(;j<argc;j++){
	if((file=fopen(argv[j],"r"))==NULL)
	  fprintf(stderr,"File [%s] does not opened.\n",argv[j]);
	else{ /* 1 file ν */
	    if((totalfilesize=getfilesize(file))<4) return Exit_errorfile;
	    rewind(file); 
	    /* magic  */
	    fread(magic,4,1,file);
	    /* getSndheader, getWAVheader,getWAVinfo Τɤ줫¹ */
	    if(strncmp(magic,"RIFF",4)==0){ /* WAV */
		if(mode==1){
#ifdef USEMALLOC
		    if((ret=getWAVinfo(file,&wd,&cl))<0)
#else
		    if((ret=getWAVinfo(file,&wd,ch))<0)
#endif /* USEMALLOC */
		      fprintf(stderr,"getWAVinfo() error.(%s)\n",argv[j]);
#ifdef USEMALLOC
		    putlongwav(argv[j],totalfilesize,&wd,cl,param);
		    Clclear(cl); cl=NULL;
#else
		    putlongwav(argv[j],totalfilesize,&wd,ch,ret,param);
#endif /* USEMALLOC */
		}
		else if(mode==0){
		    if((ret=getWAVheader(file,&wd))<0)
		      fprintf(stderr,"getWAVheader() error.(%s)\n",argv[j]);
		    if(++k==1) putdefaultheader(param);
		    putwav(argv[j],totalfilesize,&wd,param);
		}
	    }
	    else if(strncmp(magic,".snd",4)==0){ /* AU */
		if((ret=getSndheader(file,&sd))<0)
		  fprintf(stderr,"getSndheader() error.(%s)\n",argv[j]);
		if(mode==1) putlongau(argv[j],totalfilesize,&sd,param);
		else if(mode==0){
		    if(++k==1) putdefaultheader(param);
		    putau(argv[j],totalfilesize,&sd,param);
		}
	    }
	    else{ /* error */
		fprintf(stderr,"File [%s] is not supported format.\n",argv[j]);
#ifdef DEBUG1
		printf("[%c%c%c%c]\n",magic[0],magic[1],magic[2],magic[3]);
#endif
		return Exit_unsupported_format;
	    } 
	    fclose(file);
	}
    }
    return No_err;
}

/* ##### get WAV all chunk infomation ##### */
#ifdef USEMALLOC
int getWAVinfo(FILE *file,WAVdata *wd,Chunklist **Cl)
#else
int getWAVinfo(FILE *file,WAVdata *wd,Chunk *Ch)
#endif
{
    long int j;
    long int filetotalsize;
    Chunk *ch;
#ifndef USEMALLOC
    int chunks=0;
#endif

    filetotalsize=getfilesize(file); rewind(file);
    if(read4bytetolittlei(file,&j)<0 || j != sto4bytelittlei("RIFF"))
      return Err_readwavheader;
    fseek(file,4,SEEK_CUR);
    if(read4bytetolittlei(file,&j)<0 || j != sto4bytelittlei("WAVE"))
      return Err_readwavheader;
    /* chunk θ */
    while(1){
	if(ftell(file)>filetotalsize-8) break;
#ifdef USEMALLOC
	if((ch=ChnkAlloc())==NULL) return Err_malloc;
#else
	if(++chunks>Maxchunks) return Err_too_many_chunks;
	ch=&(Ch[chunks-1]);
#endif
	ch->start=ftell(file);
	fread(ch->id,4,1,file);
	if(read4bytetolittlei(file,&(ch->size))<0) return Err_readwavheader;
#ifdef USEMALLOC
	if((*Cl=addchunklist(*Cl,ch))==NULL) return Err_malloc;
#endif
	if(strncmp(ch->id,"fmt ",4)==0){ /* format chunk */
	    if(ch->size<14
	       || read2bytetolittlei(file,&(wd->fmcode))<0
	       || read2bytetolittlei(file,&(wd->channel))<0
	       || read4bytetolittlei(file,&(wd->smplrate))<0
	       || read4bytetolittlei(file,&(wd->bytepersec))<0
	       || read2bytetolittlei(file,&(wd->blocksize))<0)
	      return Err_readwavheader;
	    if(ch->size==14){ /* bitpersmpl ʤΤ support */
		if(wd->channel<=0) return Err_readwavheader;
		wd->bitpersmpl=wd->blocksize*8/wd->channel;
	    }
	    if(ch->size>=16){
		if(read2bytetolittlei(file,&(wd->bitpersmpl))<0)
		  return Err_readwavheader;
		if(ch->size>16) fseek(file,ch->size-16,SEEK_CUR);
	    }
	    else return Err_readwavheader;
	}
	else if(strncmp(ch->id,"data",4)==0){ /* data chunk */
	    wd->datalen=ch->size;
	    wd->hlen=ftell(file);
	    fseek(file,ch->size,SEEK_CUR);
	}
	else
	  fseek(file,ch->size,SEEK_CUR); /* skip */
	if((ch->size)%2)
	  fseek(file,1,SEEK_CUR); /* word alignment */
    }
#ifdef USEMALLOC
    return No_err;
#else
    return chunks;
#endif
}

/* ##### get argument ##### */
/* return value: file name start offset
*/
int getarg(int argc,char *argv[],int *Mode,int param[])
{
    int j,inoption;
    char c;

    if(argc<2) return Err_argnum;
    for(j=1;j<=allparams;j++) param[j-1]=0;
    j=1;
    if(*argv[j]=='-'){
	c=*(++argv[j]);
	while(1){
	    switch(c){
	    case 'm':
		if(*++argv[j]=='\0' && ++j>=argc) return Err_argnum;
		if((*Mode=*argv[j]-'0')<0) return Err_bad_value;
		inoption=0;
		break;
	    case 'b': param[BlockSize]=1; inoption=1; break;
	    case 't': param[DataSize]=1; inoption=1; break;
	    case 's': param[DataOffset]=1; inoption=1; break;
	    case '\0': /* return j; */
		/* stdin is not support */
	    default: return Err_unknown_option;
	    }
	    if(inoption && (c=*(++argv[j]))!='\0') continue;
	    if(++j>=argc) return Err_argnum;
	    if(*argv[j]!='-') break;
	    c=*(++argv[j]);
	}
    }
    return j;
}

#ifdef USEMALLOC
/* ##### WAV chunk list ɲ ##### */
Chunklist *addchunklist(Chunklist *cl,Chunk *ch)
{
    Chunklist *cl0;

    if(ch==NULL) return NULL;
    if(cl==NULL){
	if((cl=ChnkListAlloc())==NULL) return NULL;
	cl->p=ch;
	cl->next=NULL;
	return cl;
    }
    else{
	cl0=cl;
	while(cl->next!=NULL) cl=cl->next;
	if((cl->next=ChnkListAlloc())==NULL) return NULL;
	(cl->next)->p=ch;
	(cl->next)->next=NULL;
	return cl0;
    }
}

/* ##### Chunklist  clear ##### */
void Clclear(Chunklist *cl)
{
    Chunklist *cl0;

    if(cl==NULL) return;
    do{
	cl0=cl;
	cl=cl->next;
	free(cl0->p); free(cl0);
    }while(cl!=NULL);
}
#endif /* USEMALLOC */

/* ##### file size ֤ ##### */
long int getfilesize(FILE *file)
{
    if(file==NULL) return -1;
    fseek(file,0L,SEEK_END);
    return ftell(file);
}


/* ##### ϴؿ ##### */
#ifdef USEMALLOC
void putlongwav(char *fname,long int fsize,WAVdata *wd,Chunklist *cl,
		int param[])
#else
void putlongwav(char *fname,long int fsize,WAVdata *wd,Chunk *ch,
		int chunks,int param[])
#endif /* USEMALLOC */
{
    printf("%s : WAV, %ld Byte\n",fname,fsize);
    printf("  formatID: %04xH",wd->fmcode);
    switch(wd->fmcode){
    case 1: printf(" (linear)\n"); break;
    case 6: printf(" ( A-Law)\n"); break;
    case 7: printf(" (mu-Law)\n"); break;
    default: printf(" (unknown encoding)\n"); break;
    } 
    printf("  channel      : %d\n",wd->channel);
    printf("  samplingrate : %ld\n",wd->smplrate);
    printf("  byte/sec     : %ld\n",wd->bytepersec);
    printf("  blocksize    : %d\n",wd->blocksize);
    printf("  bit/sample   : %d\n",wd->bitpersmpl);
    if(wd->blocksize!=0)
      printf("  totalblocks  : %ld\n",wd->datalen/wd->blocksize);
    printf("  dataoffset   : %ld\n",wd->hlen);
    printf("  datasize     : %ld\n",wd->datalen);
#ifdef USEMALLOC
    putchunklist(cl);
#else
    putchunklist(ch,chunks);
#endif
    printf("\n");
}

void putlongau(char *fname,long int fsize,Snddata *sd,int param[])
{
    int blocksize,bitpersmpl;

    printf("%s : AU, %ld Byte\n",fname,fsize);
    printf("  formatID: %04lxH",sd->fmcode);
    switch(sd->fmcode){
    case 1: printf(" (mu-Law)\n"); bitpersmpl=8; break;
    case 2: printf(" (linear)\n"); bitpersmpl=8; break;
    case 3: printf(" (linear)\n"); bitpersmpl=16; break;
    case 4: printf(" (linear)\n"); bitpersmpl=24; break;
    case 5: printf(" (linear)\n"); bitpersmpl=32; break;
    default: printf(" (unknown encoding)\n"); bitpersmpl=8; break;
    }
    blocksize=sd->channel*bitpersmpl/8;
    printf("  channel      : %ld\n",sd->channel);
    printf("  samplingrate : %ld\n",sd->smplrate);
    printf("  byte/sec     : %ld\n",sd->smplrate*blocksize);
    printf("  blocksize    : %d\n",blocksize);
    printf("  bit/sample   : %d\n",bitpersmpl);
    printf("  totalblocks  : %ld\n",sd->datalen/blocksize);
    printf("  dataoffset   : %ld\n",sd->hlen);
    printf("  datasize     : %ld\n",sd->datalen);
    printf("\n");
}

#ifdef USEMALLOC
void putchunklist(Chunklist *cl)
#else
void putchunklist(Chunk *ch,int chunks)
#endif /* USEMALLOC */
{
    int j;
    Chunk *p;

    printf("  chunks :\n");
#ifdef USEMALLOC
    for(j=1;cl!=NULL;j++,cl=cl->next){
	p=cl->p;
#else
    for(j=1;j<=chunks;j++){
	p=&(ch[j-1]);
#endif /* USEMALLOC */
	printf("  [%d] \"%c%c%c%c\" (size: %ld, start: %ld)\n",j,
	       (p->id)[0],(p->id)[1],(p->id)[2],(p->id)[3],p->size,p->start);
    }
}

void putdefaultheader(int param[])
{
    printf("TYPE smplrate encode ch bit totalblocks");
    if(param[BlockSize]) printf(" bs");
    if(param[DataSize]) printf(" totaldatasize");
    if(param[DataOffset]) printf(" datastart");
    printf("   filesize filename\n");
}

void putwav(char *fname,long int fsize,WAVdata *wd,int param[])
{
    printf(" WAV %8ld",wd->smplrate);
    switch(wd->fmcode){
    case 1: printf(" linear"); break;
    case 6: printf("  A-Law"); break;
    case 7: printf(" mu-Law"); break;
    default: printf("  %04xh",wd->fmcode); break;
    }
    printf(" %2d %3d",wd->channel,wd->bitpersmpl);
    if(wd->blocksize!=0) printf(" %11ld",wd->datalen/wd->blocksize);
    else printf(" -----------");
    if(param[BlockSize]) printf(" %2d",wd->blocksize);
    if(param[DataSize]) printf(" %13ld",wd->datalen);
    if(param[DataOffset]) printf(" %9ld",wd->hlen);
    printf(" %10ld %s\n",fsize,fname);
}

void putau(char *fname,long int fsize,Snddata *sd,int param[])
{
    int bitpersmpl,blocksize;

    printf("  AU %8ld",sd->smplrate);
    switch(sd->fmcode){
    case 1: printf(" mu-Law"); bitpersmpl=8; break;
    case 2: printf(" linear"); bitpersmpl=8; break;
    case 3: printf(" linear"); bitpersmpl=16; break;
    case 4: printf(" linear"); bitpersmpl=24; break;
    case 5: printf(" linear"); bitpersmpl=32; break;
    default: printf("  %04lxh",sd->fmcode); bitpersmpl=8; break;
    }
    printf(" %2ld %3d",sd->channel,bitpersmpl);
    blocksize=sd->channel*bitpersmpl/8;
    printf(" %11ld",sd->datalen/blocksize);
    if(param[BlockSize]) printf(" %2d",blocksize);
    if(param[DataSize]) printf(" %13ld",sd->datalen);
    if(param[DataOffset]) printf(" %9ld",sd->hlen);
    printf(" %10ld %s\n",fsize,fname);
}

/* ##### usage ##### */
void usage(void)
{
    char *mess[]={
	"Usage: aufinfo (-m[mode]) (-b) (-t) (-s) [file] ([file]...)",
	" [mode] =0: ǡ (default)",
	"        =1: ܺٽ",
	" -b: 1 ֥åΥɽ (default: ʤ)",
	" -t: ǡ (=֥åߥ֥å) ɽ"
	" (default: ʤ)",
	" -s: ǡϰ֤ɽ (default: ʤ)",
	NULL
    };
    int j;

    for(j=1;mess[j-1]!=NULL;j++)
      fprintf(stderr,"%s\n",mess[j-1]);
}

